#include "rank_revealing_algorithms_intel_mkl.h"

/* column pivoted QR for mxn A 
input: matrix M
outputs: matrices Q and R and vector I such that A(:,I) = QR */
void pivotedQR_mkl(mat *M, mat **Q, mat **R, vec **I){
    int i,j,k,m,n,Rrows,Rcols,Qrows,Qcols;
    mat *Mwork;
    vec *col_vec;
    m = M->nrows; n = M->ncols;
    k = min(m,n);
    
    Mwork = matrix_new(m,n);
    matrix_copy(Mwork,M);

    int *Iarr = (int*)malloc(n*sizeof(int));
    double *tau_arr = (double*)malloc(min(m,n)*sizeof(double));

    // set up dimensions
    if(m <= n){
        Qrows = k; Qcols = k;
        Rrows = k; Rcols = n;
    } else {
        Qrows = m; Qcols = k;
        Rrows = k; Rcols = k;
    }
    
    // get R
    LAPACKE_dgeqp3(CblasColMajor, Mwork->nrows, Mwork->ncols, Mwork->d, Mwork->nrows, Iarr, tau_arr);

    *R = matrix_new(Rrows,Rcols);
    for(i=0; i<Rrows; i++){
        for(j=i; j<Rcols; j++){
            matrix_set_element(*R,i,j,matrix_get_element(Mwork,i,j));
        }
    }

    // get Q
    LAPACKE_dorgqr(CblasColMajor, Mwork->nrows, Mwork->nrows, min(Mwork->nrows,Mwork->ncols), Mwork->d, Mwork->nrows, tau_arr);

    *Q = matrix_new(Qrows,Qcols);
    for(i=0; i<Qrows; i++){
        for(j=0; j<Qcols; j++){
            matrix_set_element(*Q,i,j,matrix_get_element(Mwork,i,j));
        }
    }

    // get I
    *I = vector_new(n);
    for(i=0; i<n; i++){
        vector_set_element(*I,i,Iarr[i]-1);
    }


    // free temp variables
    matrix_delete(Mwork);
    free(Iarr);
    free(tau_arr);
}



/* generates the householder vector as a matrix for use in the 
partial column pivoted QR routine */
void get_householder_matrix(vec *x, int ind1, int ind2, mat *H){
    int i,j;
    double val,normval;

    for(i=ind1; i<ind2; i++){
        matrix_set_element(H,i,0,vector_get_element(x,i));
    }

    normval = 0;
    for(i=ind1; i<ind2; i++){
        val = vector_get_element(x,i);
        normval += val*val;
    }
    normval = sqrt(normval);

    matrix_set_element(H,ind1,0,matrix_get_element(H,ind1,0) - normval);


    val = get_matrix_frobenius_norm(H);
    if(val > 0){
        matrix_scale(H, sqrt(2/(val*val)));
    }
}


/* computes the partial pivoted QR factorization of speicifed rank 
: A(:,I) \approx Qk*Rk 
inputs: matrix M and rank k
outputs: returned rank frank, matrices Qk, Rj, index vector I
*/
void pivoted_QR_of_specified_rank(mat *M, int k, int *frank, mat **Qk, mat **Rk, vec **I){
    int i,j,ind,q,m,n,max_colnorm_index,column_norm_zero,loop_lim,checkQRresult;
    double max_colnorm,tempval,val1,val2,val3;
    int * max_colnorm_indices; 
    mat *Q, *R, *Rk1, *Rk2, *invRk1, *H, *HHt, *HtR, *HHtR, *QHHt, *T;
    mat *QH;
    vec *column_norms1, *column_norms2, *row_vec1, *row_vec2, *col_vec1, *col_vec2, *Ientries;
    m = M->nrows;
    n = M->ncols;

    // set up mats
    Q = matrix_new(m,m);
    R = matrix_new(m,n);
    H = matrix_new(m,1);
    QHHt = matrix_new(m,m);
    HtR = matrix_new(1,n);
    HHtR = matrix_new(m,n);
    QH = matrix_new(m,1);

    matrix_copy(R,M);
    
    initialize_identity_matrix(Q);
    
    column_norms1 = vector_new(n);
    column_norms2 = vector_new(n);
    compute_matrix_column_norms(M,column_norms1); 

    // loop parameters
    column_norm_zero = 0;
    *frank = 0;

    //construct vector I
    *I = vector_new(n);
    for(i=0; i<n; i++){
        vector_set_element(*I,i,i);
    }

    for(i=0; !column_norm_zero && i<k; i++){
        
        if( i%50 == 0 ){
            printf("=========> iteration %d of %d--->\n", i+1, k);
        }
        
        //printf("find max_colnorm..\n");
        max_colnorm = vector_get_element(column_norms1,i);
        max_colnorm_index = i;
        for(j=(i+1); j<n; j++){
            val1 = vector_get_element(column_norms1,j);
            if(val1 > max_colnorm){
                max_colnorm_index = j;
                max_colnorm = val1;   
            }
        }


        /* break; Qk and Rk already set for i>0 */
        if(vector_get_element(column_norms1,max_colnorm_index) == 0 && i>0){
            column_norm_zero = 1;
            printf("column norm zero detected and will break!\n");
            break;
        }
        else{
            *frank = i+1;
        }
       

        // swap I
        val1 = vector_get_element(*I,i);
        vector_set_element(*I,i,vector_get_element(*I,max_colnorm_index));
        vector_set_element(*I,max_colnorm_index,val1);


        // swap R columns
        col_vec1 = vector_new(R->nrows);
        col_vec2 = vector_new(R->nrows);
        matrix_get_col(R,i,col_vec1);
        matrix_get_col(R,max_colnorm_index,col_vec2);
        matrix_set_col(R,i,col_vec2);
        matrix_set_col(R,max_colnorm_index,col_vec1); 
        vector_delete(col_vec1); 
        vector_delete(col_vec2); 


        // swap column norms
        val1 = vector_get_element(column_norms1, max_colnorm_index);
        vector_set_element(column_norms1,max_colnorm_index,vector_get_element(column_norms1,i));
        vector_set_element(column_norms1,i,val1);


        // get H and transform R and Q
        col_vec1 = vector_new(R->nrows);
        matrix_get_col(R,i,col_vec1);
        matrix_scale(H,0);
        get_householder_matrix(col_vec1, i, m, H);
        vector_delete(col_vec1); 

        matrix_transpose_matrix_mult(H,R,HtR);
        matrix_matrix_mult(H,HtR,HHtR);
        matrix_sub(R, HHtR);

        matrix_matrix_mult(Q,H,QH);
        matrix_matrix_transpose_mult(QH,H,QHHt);
        matrix_sub(Q,QHHt);

        //downgrade norms
        if(i != (n-1)){
            #pragma omp parallel shared(column_norms1,R,i) private(ind,val1,val2) 
            {
            #pragma omp for
            for(ind = i+1; ind < n ; ind++){
                val1 = vector_get_element(column_norms1,ind);
                val2 = matrix_get_element(R,i,ind); 
                vector_set_element(column_norms1,ind,val1 - val2*val2);
            } 
            }
        }
    }        

    //construct Qk and Rk
    *Qk = matrix_new(m,*frank);
    *Rk = matrix_new(*frank,n);

    fill_matrix_from_first_columns(Q, *frank, *Qk);
    fill_matrix_from_first_rows(R, *frank, *Rk);

    
    // delete temp variables
    matrix_delete(Q);
    matrix_delete(R);
    matrix_delete(H);
    matrix_delete(QHHt);
    matrix_delete(HtR);
    matrix_delete(HHtR);
    matrix_delete(QH);
    vector_delete(column_norms1);
    vector_delete(column_norms2);
}


/* randQB single vector algorithm with power method 
inputs: matrix M, rank k [integer (>0)], power scheme parameter p [ integer (>=0) ]
outputs: matrices Q and B s.t. M \approx Q*B 
*/
void randQB_p(mat *M, int k, int p, mat **Q, mat **B){
    int i,j,m,n;
    double dotp;
    mat *A,*RN;
    vec *ej,*rj,*pj,*qj,*qi,*yj,*bj;

    m = M->nrows; n = M->ncols;

    // build random matrix
    printf("form RN..\n");
    RN = matrix_new(n, k);
    initialize_random_matrix(RN);

    // set up mats
    A = matrix_new(m,n);
    *Q = matrix_new(m,k);
    *B = matrix_new(k,n);

    ej = vector_new(k);
    rj = vector_new(n);
    pj = vector_new(n);
    yj = vector_new(m);
    qj = vector_new(m);
    qi = vector_new(m);
    bj = vector_new(n);


    // copy M to A
    matrix_copy(A,M);

    for(j=0; j<k; j++){
        if(j%20 == 0){
            printf("in iteration %d\n", j);
        }
        vector_scale(ej,0);
        vector_set_element(ej,j,1); 
        matrix_vector_mult(RN,ej,rj);
        matrix_vector_mult(A,rj,yj);

        // power method
        for(i=0; i<p; i++){
            matrix_transpose_vector_mult(A,yj,pj);
            matrix_vector_mult(A,pj,yj);
        }

        // yj = yj - Q(:,i)*dot(Q(:,i),yj);
        for(i=0; i<(j-1); i++){
            matrix_get_col(*Q, i, qi);
            dotp = vector_dot_product(qi,yj);
            vector_scale(qi,dotp);
            vector_sub(yj,qi);
        }
        vector_copy(qj,yj);
        vector_scale(qj,1.0/vector_get2norm(qj));

        matrix_set_col(*Q, j, qj);

        matrix_transpose_vector_mult(A,qj,bj);

        matrix_set_row(*B, j, bj);

        // A = A - qj*bj;
        matrix_sub_column_times_row_vector(A,qj,bj);
    }

    // clean up
    matrix_delete(A);
    matrix_delete(RN);
    vector_delete(ej);
    vector_delete(rj);
    vector_delete(yj);
    vector_delete(pj);
    vector_delete(qj);
    vector_delete(qi);
    vector_delete(bj);
}

/* randQB blocked algorithm with power method 
randQB single vector algorithm with power method 
inputs: matrix M, integer kstep (block size), integer nstep (number of blocks), power scheme parameter p [ integer (>=0) ]
outputs: matrices Q and B s.t. M \approx Q*B 
*/
void randQB_pb(mat *M, int kstep, int nstep, int p, mat **Q, mat **B){
    int i,j,m,n,l,s;
    double dotp,elapsed_time;
    int *inds_local, *inds_global;
    mat *A, *RN, *RNp, *Yp, *Qp, *Bp, *AtQp, *AtQp2, *QpBp, *Qj, *QjtQp, *QjQjtQp;
    vec *ej,*rj,*pj,*qj,*qi,*yj,*bj;
    struct timeval start_timeval, end_timeval;

    m = M->nrows; n = M->ncols;
    l = kstep*nstep;

    // build random matrix
    printf("form RN..\n");
    RN = matrix_new(n, l);
    initialize_random_matrix(RN);

    // set up mats
    A = matrix_new(m,n);
    *Q = matrix_new(m,l);
    *B = matrix_new(l,n);
    RNp = matrix_new(n,kstep); 
    Yp = matrix_new(m,kstep);
    Qp = matrix_new(m,kstep);
    Bp = matrix_new(kstep,n);
    QpBp = matrix_new(m,n);
    AtQp = matrix_new(n,kstep);
    AtQp2 = matrix_new(n,kstep);


    // copy M to A
    matrix_copy(A,M);

    for(s=0; s<nstep; s++){
        printf("in step %d\n", s);

        
        gettimeofday(&start_timeval, NULL);
        inds_local = (int*)malloc(kstep*sizeof(int));
        for(i=0; i<kstep; i++){
            inds_local[i] = kstep*s + i;
        } 
        matrix_get_selected_columns(RN, inds_local, RNp);
        matrix_matrix_mult(A,RNp,Yp);
        gettimeofday(&end_timeval, NULL);
        elapsed_time = get_seconds_frac(start_timeval,end_timeval);
        //printf("elapsed time for building Yp: %4.8f sec\n", elapsed_time);


        // power method
        gettimeofday(&start_timeval, NULL);
        for(i=0; i<p; i++){
            QR_factorization_getQ(Yp, Qp);
            matrix_transpose_matrix_mult(A,Qp,AtQp);
            QR_factorization_getQ(AtQp, AtQp2);
            matrix_matrix_mult(A,AtQp2,Yp);
        }

        // Qp = qr(Yp,0) 
        QR_factorization_getQ(Yp, Qp);
        gettimeofday(&end_timeval, NULL);
        elapsed_time = get_seconds_frac(start_timeval,end_timeval);
        //printf("elapsed time power method: %4.8f sec\n", elapsed_time);


        // project Qp away from previous Q stuff
        if(s>0){
            gettimeofday(&start_timeval, NULL);
            inds_global = (int*)malloc(s*kstep*sizeof(int));
            Qj = matrix_new(m,s*kstep);
            QjtQp = matrix_new(s*kstep,kstep);
            QjQjtQp = matrix_new(m,kstep);

            for(i=0; i<(s*kstep); i++){
                inds_global[i] = i;
            }
            matrix_get_selected_columns(*Q, inds_global, Qj);

            // Yp = Qp - Q(:,J)*(Q(:,J)'*Qp);
            matrix_transpose_matrix_mult(Qj,Qp,QjtQp);
            matrix_matrix_mult(Qj,QjtQp,QjQjtQp);
            matrix_copy(Yp,Qp);
            matrix_sub(Yp,QjQjtQp);

            // Qp = qr(Yp,0);
            QR_factorization_getQ(Yp, Qp);

            gettimeofday(&end_timeval, NULL);
            elapsed_time = get_seconds_frac(start_timeval,end_timeval);
            //printf("elapsed time for reorthogonalization: %4.8f sec\n", elapsed_time);

            free(QjQjtQp);
            free(QjtQp);
            free(Qj);
            free(inds_global);
        }


        // Bp = Qp'*A
        gettimeofday(&start_timeval, NULL);
        matrix_transpose_matrix_mult(Qp,A,Bp);
        gettimeofday(&end_timeval, NULL);
        elapsed_time = get_seconds_frac(start_timeval,end_timeval);
        //printf("elapsed time for forming Bp: %4.8f sec\n", elapsed_time);


        // A = A - Qp*Bp
        gettimeofday(&start_timeval, NULL);
        matrix_matrix_mult(Qp,Bp,QpBp);
        matrix_sub(A,QpBp); 
        gettimeofday(&end_timeval, NULL);
        elapsed_time = get_seconds_frac(start_timeval,end_timeval);
        //printf("elapsed time for updating A: %4.8f sec\n", elapsed_time);


        // Q(:,ind) = Qp; B(ind,:) = Bp;
        gettimeofday(&start_timeval, NULL);
        matrix_set_selected_columns(*Q, inds_local, Qp);
        matrix_set_selected_rows(*B, inds_local, Bp);
        gettimeofday(&end_timeval, NULL);
        elapsed_time = get_seconds_frac(start_timeval,end_timeval);
        //printf("elapsed time for updating Q and B: %4.8f sec\n", elapsed_time);
        
        free(inds_local);
    }


    // clean up
    matrix_delete(A);
    matrix_delete(RNp); 
    matrix_delete(Yp);
    matrix_delete(Qp);
    matrix_delete(Bp);
    matrix_delete(QpBp);
    matrix_delete(AtQp);
    matrix_delete(AtQp2);
}

